@@ -2,13 +2,15 @@ package m
 
 import (
 	"fmt"
+	"math"
 	"regexp"
 	"runtime/debug"
 	"time"
 
 	"github.com/alecthomas/chroma/v2"
 	log "github.com/sirupsen/logrus"
-	"github.com/walles/moar/m/linenumbers"
+	"github.com/walles/moar/m/linemetadata"
+	"github.com/walles/moar/m/reader"
 	"github.com/walles/moar/m/textstyles"
 	"github.com/walles/moar/twin"
 )
@@ -42,7 +44,8 @@ type eventMaybeDone struct{}
 
 // Pager is the main on-screen pager
 type Pager struct {
-	reader              *Reader
+	reader              *reader.ReaderImpl
+	filteringReader     FilteringReader
 	screen              twin.Screen
 	quit                bool
 	scrollPosition      scrollPosition
@@ -55,6 +58,7 @@ type Pager struct {
 
 	searchString  string
 	searchPattern *regexp.Regexp
+	filterPattern *regexp.Regexp
 
 	// We used to have a "Following" field here. If you want to follow, set
 	// TargetLineNumber to LineNumberMax() instead, see below.
@@ -81,14 +85,21 @@ type Pager struct {
 
 	SideScrollAmount int // Should be positive
 
-	// If non-nil, scroll to this line number as soon as possible. Set this
-	// value to LineNumberMax() to follow the end of the input (tail).
-	TargetLineNumber *linenumbers.LineNumber
+	// If non-nil, scroll to this line as soon as possible. Set this value to
+	// IndexMax() to follow the end of the input (tail).
+	//
+	// NOTE: Always use setTargetLine() to keep the reader in sync with the
+	// pager!
+	TargetLine *linemetadata.Index
 
 	// If true, pager will clear the screen on return. If false, pager will
 	// clear the last line, and show the cursor.
 	DeInit bool
 
+	// If DeInit is false, leave this number of lines for the shell prompt after
+	// exiting
+	DeInitFalseMargin int
+
 	WithTerminalFg bool // If true, don't set linePrefix
 
 	// Length of the longest line displayed. This is used for limiting scrolling to the right.
@@ -103,13 +114,12 @@ type Pager struct {
 }
 
 type _PreHelpState struct {
-	reader              *Reader
 	scrollPosition      scrollPosition
 	leftColumnZeroBased int
-	targetLineNumber    *linenumbers.LineNumber
+	targetLine          *linemetadata.Index
 }
 
-var _HelpReader = NewReaderFromText("Help", `
+var _HelpReader = reader.NewFromText("Help", `
 Welcome to Moar, the nice pager!
 
 Miscellaneous
@@ -137,6 +147,14 @@ Moving around
 * Half page 'u'p / 'd'own, or CTRL-u / CTRL-d
 * RETURN moves down one line
 
+Filtering
+---------
+Type '&' to start filtering, then type your filter expression.
+
+While filtering, arrow keys, PageUp, PageDown, Home and End work as usual.
+
+Press 'ESC' or RETURN to exit filtering mode.
+
 Searching
 ---------
 * Type / to start searching, then type what you want to find
@@ -163,12 +181,12 @@ Available at https://github.com/walles/moar/.
 `)
 
 // NewPager creates a new Pager with default settings
-func NewPager(r *Reader) *Pager {
+func NewPager(r *reader.ReaderImpl) *Pager {
 	var name string
-	if r == nil || r.name == nil || len(*r.name) == 0 {
+	if r == nil || r.Name == nil || len(*r.Name) == 0 {
 		name = "Pager"
 	} else {
-		name = "Pager " + *r.name
+		name = "Pager " + *r.Name
 	}
 
 	pager := Pager{
@@ -184,6 +202,10 @@ func NewPager(r *Reader) *Pager {
 	}
 
 	pager.mode = PagerModeViewing{pager: &pager}
+	pager.filteringReader = FilteringReader{
+		BackingReader: r,
+		FilterPattern: &pager.filterPattern,
+	}
 
 	return &pager
 }
@@ -198,6 +220,26 @@ func (p *Pager) visibleHeight() int {
 	return height
 }
 
+// How many cells are needed for this line number?
+//
+// Returns 0 if line numbers are disabled.
+func (p *Pager) getLineNumberPrefixLength(lineNumber linemetadata.Number) int {
+	if !p.ShowLineNumbers {
+		return 0
+	}
+
+	length := len(lineNumber.Format()) + 1 // +1 for the space after the line number
+
+	if length < 4 {
+		// 4 = space for 3 digits followed by one whitespace
+		//
+		// https://github.com/walles/moar/issues/38
+		return 4
+	}
+
+	return length
+}
+
 // Draw the footer string at the bottom using the status bar style
 func (p *Pager) setFooter(footer string) {
 	width, height := p.screen.Size()
@@ -221,10 +263,9 @@ func (p *Pager) Quit() {
 
 	// Reset help
 	p.isShowingHelp = false
-	p.reader = p.preHelpState.reader
 	p.scrollPosition = p.preHelpState.scrollPosition
 	p.leftColumnZeroBased = p.preHelpState.leftColumnZeroBased
-	p.TargetLineNumber = p.preHelpState.targetLineNumber
+	p.setTargetLine(p.preHelpState.targetLine)
 	p.preHelpState = nil
 }
 
@@ -254,27 +295,58 @@ func (p *Pager) moveRight(delta int) {
 	}
 }
 
+func (p *Pager) Reader() reader.Reader {
+	if p.isShowingHelp {
+		return _HelpReader
+	}
+	return &p.filteringReader
+}
+
 func (p *Pager) handleScrolledUp() {
-	p.TargetLineNumber = nil
+	p.setTargetLine(nil)
 }
 
 func (p *Pager) handleScrolledDown() {
 	if p.isScrolledToEnd() {
 		// Follow output
-		reallyHigh := linenumbers.LineNumberMax()
-		p.TargetLineNumber = &reallyHigh
+		reallyHigh := linemetadata.IndexMax()
+		p.setTargetLine(&reallyHigh)
 	} else {
-		p.TargetLineNumber = nil
+		p.setTargetLine(nil)
 	}
 }
 
+// Except for setting TargetLine, this method also syncs with the reader so that
+// the reader knows how many lines it needs to fetch.
+func (p *Pager) setTargetLine(targetLine *linemetadata.Index) {
+	log.Trace("Pager: Setting target line to ", targetLine, "...")
+	p.TargetLine = targetLine
+	if targetLine == nil {
+		// No target, just do your thing
+		p.reader.SetPauseAfterLines(reader.DEFAULT_PAUSE_AFTER_LINES)
+		return
+	}
+
+	// The value 1000 here is supposed to be larger than any possible screen
+	// height, to give us some lookahead and to avoid fetching too few lines.
+	targetValue := targetLine.Index() + 1000
+	if targetValue < targetLine.Index() {
+		// Overflow detected, clip to max int
+		targetValue = math.MaxInt
+	}
+	if targetValue < reader.DEFAULT_PAUSE_AFTER_LINES {
+		targetValue = reader.DEFAULT_PAUSE_AFTER_LINES
+	}
+	p.reader.SetPauseAfterLines(targetValue)
+}
+
 // StartPaging brings up the pager on screen
 func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chromaFormatter *chroma.Formatter) {
 	log.Info("Pager starting")
 
 	defer func() {
-		if p.reader.err != nil {
-			log.Warnf("Reader reported an error: %s", p.reader.err.Error())
+		if p.reader.Err != nil {
+			log.Warnf("Reader reported an error: %s", p.reader.Err.Error())
 		}
 	}()
 
@@ -286,12 +358,15 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 	p.mode = PagerModeViewing{pager: p}
 	p.marks = make(map[rune]scrollPosition)
 
+	// Make sure the reader knows how many lines we want
+	p.setTargetLine(p.TargetLine)
+
 	go func() {
 		defer func() {
-			panicHandler("StartPaging()/moreLinesAvailable", recover(), debug.Stack())
+			PanicHandler("StartPaging()/moreLinesAvailable", recover(), debug.Stack())
 		}()
 
-		for range p.reader.moreLinesAdded {
+		for range p.reader.MoreLinesAdded {
 			// Notify the main loop about the new lines so it can show them
 			screen.Events() <- eventMoreLinesAvailable{}
 
@@ -307,13 +382,13 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 
 	go func() {
 		defer func() {
-			panicHandler("StartPaging()/spinner", recover(), debug.Stack())
+			PanicHandler("StartPaging()/spinner", recover(), debug.Stack())
 		}()
 
 		// Spin the spinner as long as contents is still loading
 		spinnerFrames := [...]string{"/.\\", "-o-", "\\O/", "| |"}
 		spinnerIndex := 0
-		for !p.reader.done.Load() {
+		for !p.reader.Done.Load() {
 			screen.Events() <- eventSpinnerUpdate{spinnerFrames[spinnerIndex]}
 			spinnerIndex++
 			if spinnerIndex >= len(spinnerFrames) {
@@ -329,10 +404,10 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 
 	go func() {
 		defer func() {
-			panicHandler("StartPaging()/maybeDone", recover(), debug.Stack())
+			PanicHandler("StartPaging()/maybeDone", recover(), debug.Stack())
 		}()
 
-		for range p.reader.maybeDone {
+		for range p.reader.MaybeDone {
 			screen.Events() <- eventMaybeDone{}
 		}
 	}()
@@ -351,9 +426,9 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 			//
 			// Note that we do the slow (atomic) checks only if the fast ones (no locking
 			// required) passed
-			if p.QuitIfOneScreen && !p.isShowingHelp && p.reader.done.Load() && p.reader.highlightingDone.Load() {
+			if p.QuitIfOneScreen && !p.isShowingHelp && p.reader.Done.Load() && p.reader.HighlightingDone.Load() {
 				width, height := p.screen.Size()
-				if fitsOnOneScreen(p.reader, width, height) {
+				if fitsOnOneScreen(p.reader, width, height-p.DeInitFalseMargin) {
 					// Ref:
 					// https://github.com/walles/moar/issues/113#issuecomment-1368294132
 					p.ShowLineNumbers = false // Requires a redraw to take effect, see below
@@ -406,15 +481,15 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 			return
 
 		case eventMoreLinesAvailable:
-			if p.TargetLineNumber != nil {
+			if p.TargetLine != nil {
 				// The user wants to scroll down to a specific line number
-				if linenumbers.LineNumberFromLength(p.reader.GetLineCount()).IsBefore(*p.TargetLineNumber) {
+				if linemetadata.IndexFromLength(p.Reader().GetLineCount()).IsBefore(*p.TargetLine) {
 					// Not there yet, keep scrolling
 					p.scrollToEnd()
 				} else {
 					// We see the target, scroll to it
-					p.scrollPosition = NewScrollPositionFromLineNumber(*p.TargetLineNumber, "goToTargetLineNumber")
-					p.TargetLineNumber = nil
+					p.scrollPosition = NewScrollPositionFromIndex(*p.TargetLine, "goToTargetLine")
+					p.setTargetLine(nil)
 				}
 			}
 
@@ -434,15 +509,17 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 	}
 }
 
-func fitsOnOneScreen(reader *Reader, width int, height int) bool {
-	// One extra line to account for the status bar
-	extraLines := 1
-	if reader.GetLineCount() > height-extraLines {
+// The height parameter is the terminal height minus the height of the user's
+// shell prompt.
+//
+// This way nothing gets scrolled off screen after we exit.
+func fitsOnOneScreen(reader *reader.ReaderImpl, width int, height int) bool {
+	if reader.GetLineCount() > height {
 		return false
 	}
 
-	lines := reader.GetLines(linenumbers.LineNumberFromZeroBased(0), reader.GetLineCount())
-	for _, line := range lines.lines {
+	lines := reader.GetLines(linemetadata.Index{}, reader.GetLineCount())
+	for _, line := range lines.Lines {
 		rendered := line.HighlightedTokens(twin.StyleDefault, nil, nil).StyledRunes
 		if len(rendered) > width {
 			// This line is too long to fit on one screen line, no fit
@@ -461,15 +538,15 @@ func (p *Pager) ReprintAfterExit() error {
 	screenLinesCount := len(renderedScreenLines)
 
 	_, screenHeight := p.screen.Size()
-	screenHeightWithoutFooter := screenHeight - 1
+	screenHeightWithoutFooter := screenHeight - p.DeInitFalseMargin
 	if screenLinesCount > screenHeightWithoutFooter {
 		screenLinesCount = screenHeightWithoutFooter
 	}
 
 	if screenLinesCount > 0 {
 		p.screen.ShowNLines(screenLinesCount)
+		fmt.Println()
 	}
-	fmt.Println()
 
 	return nil
 }
