@@ -12,10 +12,13 @@ type PagerModeViewing struct {
 func (m PagerModeViewing) drawFooter(statusText string, spinner string) {
 	helpText := "Press 'ESC' / 'q' to exit, '/' to search, '&' to filter, 'h' for help"
 	if m.pager.isShowingHelp {
-		helpText = "Press 'ESC' / 'q' to exit help, '/' to search, '&' to filter"
+		helpText = "Press 'ESC' / 'q' to exit help, '/' to search"
 	}
 
 	if m.pager.ShowStatusBar {
+		if len(spinner) > 0 {
+			spinner = "  " + spinner
+		}
 		m.pager.setFooter(statusText + spinner + "  " + helpText)
 	}
 }
@@ -85,15 +88,13 @@ func (m PagerModeViewing) onRune(char rune) {
 		}
 
 		p.preHelpState = &_PreHelpState{
-			reader:              p.reader,
 			scrollPosition:      p.scrollPosition,
 			leftColumnZeroBased: p.leftColumnZeroBased,
 			targetLine:          p.TargetLine,
 		}
-		p.reader = _HelpReader
 		p.scrollPosition = newScrollPosition("Pager scroll position")
 		p.leftColumnZeroBased = 0
-		p.TargetLine = nil
+		p.setTargetLine(nil)
 		p.isShowingHelp = true
 
 	case '=':
@@ -142,25 +143,29 @@ func (m PagerModeViewing) onRune(char rune) {
 
 	case '/':
 		p.mode = PagerModeSearch{pager: p, direction: SearchDirectionForward, initialScrollPosition: p.scrollPosition}
-		p.TargetLine = nil
+		p.setTargetLine(nil)
 		p.searchString = ""
 		p.searchPattern = nil
 
 	case '?':
 		p.mode = PagerModeSearch{pager: p, direction: SearchDirectionBackward, initialScrollPosition: p.scrollPosition}
-		p.TargetLine = nil
+		p.setTargetLine(nil)
 		p.searchString = ""
 		p.searchPattern = nil
 
 	case '&':
-		p.mode = &PagerModeFilter{pager: p}
-		p.searchString = ""
-		p.searchPattern = nil
-		p.filterPattern = nil
+		if !p.isShowingHelp {
+			// Filtering the help text is not supported. Feel free to work on
+			// that if you feel that's time well spent.
+			p.mode = &PagerModeFilter{pager: p}
+			p.searchString = ""
+			p.searchPattern = nil
+			p.filterPattern = nil
+		}
 
 	case 'g':
 		p.mode = &PagerModeGotoLine{pager: p}
-		p.TargetLine = nil
+		p.setTargetLine(nil)
 
 	// Should match the pagermode-not-found.go previous-search-hit bindings
 	case 'n':
@@ -172,11 +177,11 @@ func (m PagerModeViewing) onRune(char rune) {
 
 	case 'm':
 		p.mode = PagerModeMark{pager: p}
-		p.TargetLine = nil
+		p.setTargetLine(nil)
 
 	case '\'':
 		p.mode = PagerModeJumpToMark{pager: p}
-		p.TargetLine = nil
+		p.setTargetLine(nil)
 
 	case 'w':
 		p.WrapLongLines = !p.WrapLongLines
