@@ -3,20 +3,16 @@ package m
 import (
 	"fmt"
 
-	"github.com/walles/moar/m/linenumbers"
+	"github.com/walles/moar/m/linemetadata"
+	"github.com/walles/moar/m/reader"
 	"github.com/walles/moar/m/textstyles"
 	"github.com/walles/moar/twin"
 )
 
-type overflowState bool
-
-const (
-	didFit      overflowState = false
-	didOverflow overflowState = true
-)
-
 type renderedLine struct {
-	inputLine linenumbers.LineNumber
+	// Certain lines are available for viewing. This index is the (zero based)
+	// position of this line among those.
+	inputLineIndex linemetadata.Index
 
 	// If an input line has been wrapped into two, the part on the second line
 	// will have a wrapIndex of 1.
@@ -33,13 +29,13 @@ type renderedLine struct {
 
 // Refresh the whole pager display, both contents lines and the status line at
 // the bottom
-func (p *Pager) redraw(spinner string) overflowState {
+func (p *Pager) redraw(spinner string) {
 	p.screen.Clear()
 	p.longestLineLength = 0
 
 	lastUpdatedScreenLineNumber := -1
 	var renderedScreenLines [][]twin.StyledRune
-	renderedScreenLines, statusText, overflow := p.renderScreenLines()
+	renderedScreenLines, statusText := p.renderScreenLines()
 	for screenLineNumber, row := range renderedScreenLines {
 		lastUpdatedScreenLineNumber = screenLineNumber
 		column := 0
@@ -55,7 +51,7 @@ func (p *Pager) redraw(spinner string) overflowState {
 		// This happens when we're done
 		eofSpinner = "---"
 	}
-	spinnerLine := textstyles.StyledRunesFromString("", _EofMarkerFormat+eofSpinner, nil).StyledRunes
+	spinnerLine := textstyles.StyledRunesFromString(statusbarStyle, eofSpinner, nil).StyledRunes
 	column := 0
 	for _, cell := range spinnerLine {
 		column += p.screen.SetCell(column, lastUpdatedScreenLineNumber+1, cell)
@@ -64,7 +60,6 @@ func (p *Pager) redraw(spinner string) overflowState {
 	p.mode.drawFooter(statusText, spinner)
 
 	p.screen.Show()
-	return overflow
 }
 
 // Render screen lines into an array of lines consisting of Cells.
@@ -73,8 +68,8 @@ func (p *Pager) redraw(spinner string) overflowState {
 //
 // The lines returned by this method are decorated with horizontal scroll
 // markers and line numbers and are ready to be output to the screen.
-func (p *Pager) renderScreenLines() (lines [][]twin.StyledRune, statusText string, overflow overflowState) {
-	renderedLines, statusText, overflow := p.renderLines()
+func (p *Pager) renderScreenLines() (lines [][]twin.StyledRune, statusText string) {
+	renderedLines, statusText := p.renderLines()
 	if len(renderedLines) == 0 {
 		return
 	}
@@ -96,7 +91,7 @@ func (p *Pager) renderScreenLines() (lines [][]twin.StyledRune, statusText strin
 		}
 	}
 
-	return screenLines, statusText, overflow
+	return screenLines, statusText
 }
 
 // Render all lines that should go on the screen.
@@ -109,49 +104,26 @@ func (p *Pager) renderScreenLines() (lines [][]twin.StyledRune, statusText strin
 // The maximum number of lines returned by this method is limited by the screen
 // height. If the status line is visible, you'll get at most one less than the
 // screen height from this method.
-func (p *Pager) renderLines() ([]renderedLine, string, overflowState) {
-	wantedLineCount := p.visibleHeight()
-
-	screenOverflow := didFit
-
-	var lineNumber linenumbers.LineNumber
-	if p.lineNumber() != nil {
-		lineNumber = *p.lineNumber()
-	} else {
-		// No lines to show, line number doesn't matter, pick anything. But we
-		// still want one so that we can get the status text from the reader
-		// below.
-		lineNumber = linenumbers.LineNumber{}
+func (p *Pager) renderLines() ([]renderedLine, string) {
+	var lineIndex linemetadata.Index
+	if p.lineIndex() != nil {
+		lineIndex = *p.lineIndex()
 	}
-
-	if !lineNumber.IsZero() {
-		// We're scrolled down, meaning everything is not visible on screen
-		screenOverflow = didOverflow
-	}
-
-	inputLines, readerOverflow := p.reader.GetLines(lineNumber, wantedLineCount)
-	if inputLines.lines == nil {
+	inputLines := p.Reader().GetLines(lineIndex, p.visibleHeight())
+	if len(inputLines.Lines) == 0 {
 		// Empty input, empty output
-		return []renderedLine{}, inputLines.statusText, didFit
-	}
-	if readerOverflow == didOverflow {
-		// This is not the whole input
-		screenOverflow = didOverflow
+		return []renderedLine{}, inputLines.StatusText
 	}
 
+	lastVisibleLineNumber := inputLines.Lines[len(inputLines.Lines)-1].Number
+	numberPrefixLength := p.getLineNumberPrefixLength(lastVisibleLineNumber)
+
 	allLines := make([]renderedLine, 0)
-	for lineIndex, line := range inputLines.lines {
-
-		lineNumber := inputLines.firstLine.NonWrappingAdd(lineIndex)
-
-		rendering, lineOverflow := p.renderLine(line, lineNumber, p.scrollPosition.internalDontTouch)
-		if lineOverflow == didOverflow {
-			// Everything did not fit
-			screenOverflow = didOverflow
-		}
+	for _, line := range inputLines.Lines {
+		rendering := p.renderLine(line, numberPrefixLength)
 
 		var onScreenLength int
-		for i := 0; i < len(rendering); i++ {
+		for i := range rendering {
 			trimmedLen := len(twin.TrimSpaceRight(rendering[i].cells))
 			if trimmedLen > onScreenLength {
 				onScreenLength = trimmedLen
@@ -176,12 +148,12 @@ func (p *Pager) renderLines() ([]renderedLine, string, overflowState) {
 	// screen
 	firstVisibleIndex := -1 // Not found
 	for index, line := range allLines {
-		if p.lineNumber() == nil {
+		if p.lineIndex() == nil {
 			// Expected zero lines but got some anyway, grab the first one!
 			firstVisibleIndex = index
 			break
 		}
-		if line.inputLine == *p.lineNumber() && line.wrapIndex == p.deltaScreenLines() {
+		if line.inputLineIndex == *p.lineIndex() && line.wrapIndex == p.deltaScreenLines() {
 			firstVisibleIndex = index
 			break
 		}
@@ -190,21 +162,17 @@ func (p *Pager) renderLines() ([]renderedLine, string, overflowState) {
 		panic(fmt.Errorf("scrollPosition %#v not found in allLines size %d",
 			p.scrollPosition, len(allLines)))
 	}
-	if firstVisibleIndex != 0 {
-		// We're scrolled down, meaning everything is not visible on screen
-		screenOverflow = didOverflow
-	}
 
 	// Drop the lines that should go above the screen
 	allLines = allLines[firstVisibleIndex:]
 
+	wantedLineCount := p.visibleHeight()
 	if len(allLines) <= wantedLineCount {
 		// Screen has enough room for everything, return everything
-		return allLines, inputLines.statusText, screenOverflow
+		return allLines, inputLines.StatusText
 	}
 
-	screenOverflow = didOverflow
-	return allLines[0:wantedLineCount], inputLines.statusText, screenOverflow
+	return allLines[0:wantedLineCount], inputLines.StatusText
 }
 
 // Render one input line into one or more screen lines.
@@ -214,38 +182,31 @@ func (p *Pager) renderLines() ([]renderedLine, string, overflowState) {
 //
 // lineNumber and numberPrefixLength are required for knowing how much to
 // indent, and to (optionally) render the line number.
-func (p *Pager) renderLine(line *Line, lineNumber linenumbers.LineNumber, scrollPosition scrollPositionInternal) ([]renderedLine, overflowState) {
-	highlighted := line.HighlightedTokens(p.linePrefix, p.searchPattern, &lineNumber)
+func (p *Pager) renderLine(line *reader.NumberedLine, numberPrefixLength int) []renderedLine {
+	highlighted := line.HighlightedTokens(plainTextStyle, standoutStyle, p.searchPattern)
 	var wrapped [][]twin.StyledRune
-	overflow := didFit
 	if p.WrapLongLines {
 		width, _ := p.screen.Size()
-		wrapped = wrapLine(width-numberPrefixLength(p, scrollPosition), highlighted.StyledRunes)
+		wrapped = wrapLine(width-numberPrefixLength, highlighted.StyledRunes)
 	} else {
 		// All on one line
 		wrapped = [][]twin.StyledRune{highlighted.StyledRunes}
 	}
 
-	if len(wrapped) > 1 {
-		overflow = didOverflow
-	}
-
 	rendered := make([]renderedLine, 0)
 	for wrapIndex, inputLinePart := range wrapped {
+		lineNumber := line.Number
 		visibleLineNumber := &lineNumber
 		if wrapIndex > 0 {
 			visibleLineNumber = nil
 		}
 
-		decorated, localOverflow := p.decorateLine(visibleLineNumber, inputLinePart, scrollPosition)
-		if localOverflow == didOverflow {
-			overflow = didOverflow
-		}
+		decorated := p.decorateLine(visibleLineNumber, numberPrefixLength, inputLinePart)
 
 		rendered = append(rendered, renderedLine{
-			inputLine: lineNumber,
-			wrapIndex: wrapIndex,
-			cells:     decorated,
+			inputLineIndex: line.Index,
+			wrapIndex:      wrapIndex,
+			cells:          decorated,
 		})
 	}
 
@@ -255,19 +216,17 @@ func (p *Pager) renderLine(line *Line, lineNumber linenumbers.LineNumber, scroll
 		rendered[len(rendered)-1].trailer = highlighted.Trailer
 	}
 
-	return rendered, overflow
+	return rendered
 }
 
 // Take a rendered line and decorate as needed:
 //   - Line number, or leading whitespace for wrapped lines
 //   - Scroll left indicator
 //   - Scroll right indicator
-func (p *Pager) decorateLine(lineNumberToShow *linenumbers.LineNumber, contents []twin.StyledRune, scrollPosition scrollPositionInternal) ([]twin.StyledRune, overflowState) {
+func (p *Pager) decorateLine(lineNumberToShow *linemetadata.Number, numberPrefixLength int, contents []twin.StyledRune) []twin.StyledRune {
 	width, _ := p.screen.Size()
 	newLine := make([]twin.StyledRune, 0, width)
-	numberPrefixLength := numberPrefixLength(p, scrollPosition)
 	newLine = append(newLine, createLinePrefix(lineNumberToShow, numberPrefixLength)...)
-	overflow := didFit
 
 	// Find the first and last fully visible runes.
 	var firstVisibleRuneIndex *int
@@ -352,9 +311,6 @@ func (p *Pager) decorateLine(lineNumberToShow *linenumbers.LineNumber, contents
 
 		// Set can-scroll-left marker
 		newLine[0] = p.ScrollLeftHint
-
-		// We're scrolled right, meaning everything is not visible on screen
-		overflow = didOverflow
 	}
 
 	// Add scroll right indicator
@@ -369,18 +325,15 @@ func (p *Pager) decorateLine(lineNumberToShow *linenumbers.LineNumber, contents
 		}
 
 		newLine[len(newLine)-1] = p.ScrollRightHint
-
-		// Some text is out of bounds to the right
-		overflow = didOverflow
 	}
 
-	return newLine, overflow
+	return newLine
 }
 
 // Generate a line number prefix of the given length.
 //
 // Can be empty or all-whitespace depending on parameters.
-func createLinePrefix(lineNumber *linenumbers.LineNumber, numberPrefixLength int) []twin.StyledRune {
+func createLinePrefix(lineNumber *linemetadata.Number, numberPrefixLength int) []twin.StyledRune {
 	if numberPrefixLength == 0 {
 		return []twin.StyledRune{}
 	}
