@@ -9,6 +9,7 @@ import (
 	"runtime/debug"
 	"strconv"
 	"strings"
+	"sync"
 	"time"
 
 	"github.com/alecthomas/chroma/v2"
@@ -19,7 +20,8 @@ import (
 	"golang.org/x/term"
 
 	"github.com/walles/moar/m"
-	"github.com/walles/moar/m/linenumbers"
+	"github.com/walles/moar/m/linemetadata"
+	"github.com/walles/moar/m/reader"
 	"github.com/walles/moar/m/textstyles"
 	"github.com/walles/moar/twin"
 )
@@ -135,8 +137,8 @@ func parseUnprintableStyle(styleOption string) (textstyles.UnprintableStyleT, er
 
 func parseScrollHint(scrollHint string) (twin.StyledRune, error) {
 	scrollHint = strings.ReplaceAll(scrollHint, "ESC", "\x1b")
-	hintAsLine := m.NewLine(scrollHint)
-	parsedTokens := hintAsLine.HighlightedTokens(twin.StyleDefault, nil, nil).StyledRunes
+	hintAsLine := reader.NewLine(scrollHint)
+	parsedTokens := hintAsLine.HighlightedTokens(twin.StyleDefault, nil, nil, nil).StyledRunes
 	if len(parsedTokens) == 1 {
 		return parsedTokens[0], nil
 	}
@@ -177,7 +179,7 @@ func pumpToStdout(inputFilenames ...string) error {
 		// If we get both redirected stdin and an input filenames, should only
 		// copy the files and ignore stdin, because that's how less works.
 		for _, inputFilename := range inputFilenames {
-			inputFile, _, err := m.ZOpen(inputFilename)
+			inputFile, _, err := reader.ZOpen(inputFilename)
 			if err != nil {
 				return fmt.Errorf("Failed to open %s: %w", inputFilename, err)
 			}
@@ -199,37 +201,11 @@ func pumpToStdout(inputFilenames ...string) error {
 	return nil
 }
 
-// Duplicate of m/reader.go:tryOpen
-func tryOpen(filename string) error {
-	// Try opening the file
-	tryMe, err := os.Open(filename)
-	if err != nil {
-		return err
-	}
-
-	// Try reading a byte
-	buffer := make([]byte, 1)
-	_, err = tryMe.Read(buffer)
-
-	if err != nil && err.Error() == "EOF" {
-		// Empty file, this is fine
-		err = nil
-	}
-
-	closeErr := tryMe.Close()
-	if err == nil && closeErr != nil {
-		// Everything worked up until Close(), report the Close() error
-		return closeErr
-	}
-
-	return err
-}
-
 // Parses an argument like "+123" anywhere on the command line into a one-based
 // line number, and returns the remaining args.
 //
 // Returns nil on no target line number specified.
-func getTargetLineNumber(args []string) (*linenumbers.LineNumber, []string) {
+func getTargetLine(args []string) (*linemetadata.Index, []string) {
 	for i, arg := range args {
 		if !strings.HasPrefix(arg, "+") {
 			continue
@@ -252,7 +228,7 @@ func getTargetLineNumber(args []string) (*linenumbers.LineNumber, []string) {
 		remainingArgs = append(remainingArgs, args[:i]...)
 		remainingArgs = append(remainingArgs, args[i+1:]...)
 
-		returnMe := linenumbers.LineNumberFromOneBased(int(lineNumber))
+		returnMe := linemetadata.IndexFromOneBased(int(lineNumber))
 		return &returnMe, remainingArgs
 	}
 
@@ -358,8 +334,10 @@ func pagerFromArgs(
 	noStatusBar := flagSet.Bool("no-statusbar", false, "Hide the status bar, toggle with '='")
 	reFormat := flagSet.Bool("reformat", false, "Reformat some input files (JSON)")
 	flagSet.Bool("no-reformat", true, "No effect, kept for compatibility. See --reformat")
-	quitIfOneScreen := flagSet.Bool("quit-if-one-screen", false, "Don't page if contents fits on one screen")
+	quitIfOneScreen := flagSet.Bool("quit-if-one-screen", false, "Don't page if contents fits on one screen. Affected by --no-clear-on-exit-margin.")
 	noClearOnExit := flagSet.Bool("no-clear-on-exit", false, "Retain screen contents when exiting moar")
+	noClearOnExitMargin := flagSet.Int("no-clear-on-exit-margin", 1,
+		"Number of lines to leave for your shell prompt, defaults to 1")
 	statusBarStyle := flagSetFunc(flagSet, "statusbar", m.STATUSBAR_STYLE_INVERSE,
 		"Status bar `style`: inverse, plain or bold", parseStatusBarStyle)
 	unprintableStyle := flagSetFunc(flagSet, "render-unprintable", textstyles.UnprintableStyleHighlight,
@@ -388,9 +366,16 @@ func pagerFromArgs(
 		flags = append(strings.Fields(moarEnv), flags...)
 	}
 
-	targetLineNumber, remainingArgs := getTargetLineNumber(flags)
+	targetLine, remainingArgs := getTargetLine(flags)
 
 	err = flagSet.Parse(remainingArgs)
+
+	if err == nil {
+		if *noClearOnExitMargin < 0 {
+			err = fmt.Errorf("Invalid --no-clear-on-exit-margin %d, must be 0 or higher", *noClearOnExitMargin)
+		}
+	}
+
 	if err != nil {
 		if err == flag.ErrHelp {
 			printUsage(flagSet, *terminalColorsCount)
@@ -441,7 +426,7 @@ func pagerFromArgs(
 	for _, inputFilename := range flagSet.Args() {
 		// Need to check before newScreen() below, otherwise the screen
 		// will be cleared before we print the "No such file" error.
-		err := tryOpen(inputFilename)
+		err := reader.TryOpen(inputFilename)
 		if err != nil {
 			return nil, nil, chroma.Style{}, nil, logsRequested, err
 		}
@@ -480,19 +465,20 @@ func pagerFromArgs(
 	}
 
 	formatter := formatters.TTY256
-	if *terminalColorsCount == twin.ColorCount8 {
+	switch *terminalColorsCount {
+	case twin.ColorCount8:
 		formatter = formatters.TTY8
-	} else if *terminalColorsCount == twin.ColorCount16 {
+	case twin.ColorCount16:
 		formatter = formatters.TTY16
-	} else if *terminalColorsCount == twin.ColorCount24bit {
+	case twin.ColorCount24bit:
 		formatter = formatters.TTY16m
 	}
 
-	var reader *m.Reader
+	var readerImpl *reader.ReaderImpl
 	shouldFormat := *reFormat
 	if stdinIsRedirected {
 		// Display input pipe contents
-		reader, err = m.NewReaderFromStream("", os.Stdin, formatter, m.ReaderOptions{Lexer: *lexer, ShouldFormat: shouldFormat})
+		readerImpl, err = reader.NewFromStream("", os.Stdin, formatter, reader.ReaderOptions{Lexer: *lexer, ShouldFormat: shouldFormat})
 		if err != nil {
 			return nil, nil, chroma.Style{}, nil, logsRequested, err
 		}
@@ -501,7 +487,7 @@ func pagerFromArgs(
 		if len(flagSet.Args()) != 1 {
 			panic("Invariant broken: Expected exactly one filename")
 		}
-		reader, err = m.NewReaderFromFilename(flagSet.Args()[0], formatter, m.ReaderOptions{Lexer: *lexer, ShouldFormat: shouldFormat})
+		readerImpl, err = reader.NewFromFilename(flagSet.Args()[0], formatter, reader.ReaderOptions{Lexer: *lexer, ShouldFormat: shouldFormat})
 		if err != nil {
 			return nil, nil, chroma.Style{}, nil, logsRequested, err
 		}
@@ -509,7 +495,7 @@ func pagerFromArgs(
 
 	// If the user is doing "sudo something | moar" we can't show the UI until
 	// we start getting data, otherwise we'll mess up sudo's password prompt.
-	reader.AwaitFirstByte()
+	readerImpl.AwaitFirstByte()
 
 	// We got the first byte, this means sudo is done (if it was used) and we
 	// can set up the UI.
@@ -518,12 +504,12 @@ func pagerFromArgs(
 		// Ref: https://github.com/walles/moar/issues/149
 		log.Info("Failed to set up screen for paging, pumping to stdout instead: ", err)
 
-		reader.PumpToStdout()
+		readerImpl.PumpToStdout()
 
 		return nil, nil, chroma.Style{}, nil, logsRequested, nil
 	}
 
-	var style chroma.Style = *styles.Get(defaultDarkTheme)
+	var style = *styles.Get(defaultDarkTheme)
 	if *styleOption == nil {
 		t0 := time.Now()
 		screen.RequestTerminalBackgroundColor()
@@ -558,13 +544,14 @@ func pagerFromArgs(
 		style = **styleOption
 	}
 	log.Debug("Using style <", style.Name, ">")
-	reader.SetStyleForHighlighting(style)
+	readerImpl.SetStyleForHighlighting(style)
 
-	pager := m.NewPager(reader)
+	pager := m.NewPager(readerImpl)
 	pager.WrapLongLines = *wrap
 	pager.ShowLineNumbers = !*noLineNumbers
 	pager.ShowStatusBar = !*noStatusBar
 	pager.DeInit = !*noClearOnExit
+	pager.DeInitFalseMargin = *noClearOnExitMargin
 	pager.QuitIfOneScreen = *quitIfOneScreen
 	pager.StatusBarStyle = *statusBarStyle
 	pager.UnprintableStyle = *unprintableStyle
@@ -573,17 +560,36 @@ func pagerFromArgs(
 	pager.ScrollRightHint = *scrollRightHint
 	pager.SideScrollAmount = int(*shift)
 
-	pager.TargetLineNumber = targetLineNumber
-	if *follow && pager.TargetLineNumber == nil {
-		reallyHigh := linenumbers.LineNumberMax()
-		pager.TargetLineNumber = &reallyHigh
+	pager.TargetLine = targetLine
+	if *follow && pager.TargetLine == nil {
+		reallyHigh := linemetadata.IndexMax()
+		pager.TargetLine = &reallyHigh
 	}
 
 	return pager, screen, style, &formatter, logsRequested, nil
 }
 
+type logWriter struct {
+	lock   sync.Mutex
+	buffer strings.Builder
+}
+
+func (lw *logWriter) Write(p []byte) (n int, err error) {
+	lw.lock.Lock()
+	defer lw.lock.Unlock()
+
+	return lw.buffer.Write(p)
+}
+
+func (lw *logWriter) String() string {
+	lw.lock.Lock()
+	defer lw.lock.Unlock()
+
+	return lw.buffer.String()
+}
+
 func main() {
-	var loglines strings.Builder
+	var loglines logWriter
 	logsRequested := false
 	log.SetOutput(&loglines)
 	russiaNotSupported()
